VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsGIF"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

' clsGIF.cls - An LZW decoder for GIF
' Copyright (C) 1987, by Steven A. Bennett
' Ported to VB by A. Skrobov, 2004
'
' Permission is given by the author to freely redistribute and include
' this code in any program as long as this credit is given where due.
'
' In accordance with the above, I want to credit Steve Wilhite who wrote
' the code which this is heavily inspired by...
'
' GIF and 'Graphics Interchange Format' are trademarks (tm) of
' Compuserve, Incorporated, an H&R Block Company.
'
' Release Notes: This file contains a decoder routine for GIF images
' which is similar, structurally, to the original routine by Steve Wilhite.
' It is, however, somewhat noticably faster in most cases.

Private Type gif_header
    giftype As String * 6
    Width As Integer
    Height As Integer
    flags As Byte
    bgcolor As Byte
    reserved As Byte
End Type

Private Type gif_image
    left As Integer
    top As Integer
    Width As Integer
    Height As Integer
    flags As Byte
End Type

Private Type gif_control
    size As Byte
    flags As Byte
    delay As Integer
    xprntcolor As Byte
    terminator As Byte
End Type

Private Const MAX_CODES = 4095

' Static variables
Private curr_size As Integer    ' The current code size
Private clear As Integer        ' Value for a clear code
Private ending As Integer       ' Value for a ending code
Private newcodes As Integer     ' First available code
Private top_slot As Integer     ' Highest code for current size
Private slot As Integer         ' Last read code

' The following static variables are used
' for separating out codes
Private navail_bytes As Integer ' # bytes left in block
Private nbits_left As Integer   ' # bits left in current byte
Private b1 As Byte              ' Current byte
Private gif_byte_buff(0 To 256 + 3) As Byte ' Current block
Private pbytes As Integer       ' Pointer to next byte in block

Private code_mask(0 To 12) As Integer

'Following are implementation-specific globals
Private powers_of_two(0 To 12) As Integer

Private data() As Byte, pdata As Long

Private Type RGBTRIPLE
     rgbRed As Byte
     rgbGreen As Byte
     rgbBlue As Byte
End Type

Private Type BITMAPINFOHEADER
     biSize As Long
     biWidth As Long
     biHeight As Long
     biPlanes As Integer
     biBitCount As Integer
     biCompression As Long
     biSizeImage As Long
     biXPelsPerMeter As Long
     biYPelsPerMeter As Long
     biClrUsed As Long
     biClrImportant As Long
End Type
Private Type RGBQUAD
     rgbBlue As Byte
     rgbGreen As Byte
     rgbRed As Byte
     rgbReserved As Byte
End Type
'Private Type BITMAPINFO256
'     bmiHeader As BITMAPINFOHEADER
'     bmiColors(0 To 255) As RGBQUAD
'End Type
'Private Type BITMAPINFO2
'     bmiHeader As BITMAPINFOHEADER
'     bmiColor0 As Long
'     bmiColor1 As Long
'End Type
Private Const CBM_INIT = 4&
Private Declare Function CreateDIBitmap Lib "gdi32" (ByVal hDC As Long, lpbmih As BITMAPINFOHEADER, ByVal fdwInit As Long, lpbInit As Any, lpbmi As BITMAPINFOHEADER, ByVal fuUsage As Long) As Long
'Private Declare Function CreateDIBSection Lib "gdi32" (ByVal hDC As Long, pbmi As BITMAPINFOHEADER, ByVal iUsage As Long, ppvBits As Long, ByVal hSection As Long, ByVal dwOffset As Long) As Long
'Private Declare Function DeleteObject Lib "gdi32" (ByVal hObject As Long) As Long
Private Declare Function GetDC Lib "user32" (ByVal hWnd As Long) As Long
Private Declare Function ReleaseDC Lib "user32" (ByVal hWnd As Long, ByVal hDC As Long) As Long

Private Type Guid
    Data1 As Long
    Data2 As Integer
    Data3 As Integer
    Data4(0 To 7) As Byte
End Type
Private Type PICTDESC
    size As Long
    Type As Long
    hBmp As Long
    hPal As Long
    reserved As Long
End Type
Private Declare Function OleCreatePictureIndirect Lib "olepro32" (lpPictDesc As PICTDESC, riid As Guid, ByVal fOwn As Long, lplpvObj As IPicture) As Long

'Private Type SAFEARRAY2D
'    cDims As Integer
'    fFeatures As Integer
'    cbElements As Long
'    cLocks As Long
'    pvData As Long
'    cElements0 As Long
'    lLbound0 As Long
'    cElements1 As Long
'    lLbound1 As Long
'End Type
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (Dest As Any, Src As Any, ByVal Length As Long)
'Private Declare Sub FillMemory Lib "ntdll" Alias "RtlFillMemoryUlong" (Dest As Any, ByVal Length As Long, ByVal Data As Long)
'Private Declare Sub FillMemory Lib "kernel32" Alias "RtlFillMemory" (Dest As Any, ByVal Length As Long, ByVal Data As Byte)
Private Declare Sub ZeroMemory Lib "kernel32" Alias "RtlZeroMemory" (Dest As Any, ByVal Length As Long)
Private Declare Sub PutMem4 Lib "msvbvm60" (Dest As Any, ByVal data As Long)
Private Declare Sub GetMem4 Lib "msvbvm60" (Src As Any, Dest As Any)
Private Declare Function VarPtrArray Lib "msvbvm60" Alias "VarPtr" (Ary() As Any) As Long
Private Palette() As RGBTRIPLE, BitmapData() As RGBQUAD, MaskData() As Long
Private LineNo As Long, LastLine As Long, LineOff As Long
Private bih As BITMAPINFOHEADER

Private GlobalColors As Integer, GlobalPalette(0 To 255) As RGBTRIPLE
Private XprntIndex As Integer, Zombie As Boolean

Public DelayMS As Integer, NextMask As StdPicture
Public Width As Long, Height As Long

'   - This implementation-specific function is expected to return
' the next byte from the GIF file
Private Function GetByte() As Byte
    GetByte = data(pdata): pdata = pdata + 1
End Function

'   - This function takes a full line of pixels (one byte per pixel) and
' displays them (or does whatever your program wants with them...).
Private Function OutLine(pixels() As Byte, ByVal linelen As Integer) As Boolean
'assumed: LBound(pixels) = 0
Dim i As Long, x As Long, c As RGBTRIPLE, Index As Byte
    x = LineOff
    For i = 0 To linelen - 1
        Index = pixels(i)
        If Index = XprntIndex Then
'            MaskData(X, LineNo) = 0
        Else
            MaskData(x, LineNo) = vbWhite
            c = Palette(Index)
            With BitmapData(x, LineNo)
                .rgbRed = c.rgbRed
                .rgbGreen = c.rgbGreen
                .rgbBlue = c.rgbBlue
            End With
        End If
        x = x + 1
    Next
    LineNo = LineNo + 1
OutLine = (LineNo > LastLine)
End Function

' This function initializes the decoder for reading a new image.
Private Sub init_exp(ByVal size As Integer)
Dim i As Byte
powers_of_two(0) = 1
For i = 1 To 12
    powers_of_two(i) = powers_of_two(i - 1) * 2
    code_mask(i) = powers_of_two(i) - 1
Next

   curr_size = size + 1
   top_slot = powers_of_two(curr_size)
   clear = powers_of_two(size)
   ending = clear + 1
   slot = ending + 1: newcodes = ending + 1
   navail_bytes = 0: nbits_left = 0
End Sub

' - gets the next code from the GIF file.  Returns the code.
Private Function get_next_code() As Integer
Dim i As Integer, x As Integer
Dim ret As Long

   If nbits_left = 0 Then
      If navail_bytes <= 0 Then

         ' Out of bytes in current block, so read next block
         pbytes = 0
         navail_bytes = GetByte
         If navail_bytes > 0 Then
            For i = 0 To navail_bytes - 1
               x = GetByte
               gif_byte_buff(i) = x
            Next
         End If
      End If
      b1 = gif_byte_buff(pbytes): pbytes = pbytes + 1
      nbits_left = 8
      navail_bytes = navail_bytes - 1
   End If

   ret = b1 \ powers_of_two(8 - nbits_left)
   While curr_size > nbits_left
      If navail_bytes <= 0 Then

         ' Out of bytes in current block, so read next block
         pbytes = 0
         navail_bytes = GetByte
         If navail_bytes > 0 Then
            For i = 0 To navail_bytes - 1
               x = GetByte
               gif_byte_buff(i) = x
            Next
         End If
      End If
      b1 = gif_byte_buff(pbytes): pbytes = pbytes + 1
      ret = ret Or (CLng(b1) * powers_of_two(nbits_left))
      nbits_left = nbits_left + 8
      navail_bytes = navail_bytes - 1
   Wend
   nbits_left = nbits_left - curr_size
   ret = ret And code_mask(curr_size)
   get_next_code = ret
End Function


' - This function decodes an LZW image, according to the method used
' in the GIF spec.  Every *linewidth* 'characters' (ie. pixels) decoded
' will generate a call to OutLine(), which is a user specific function
' to display a line of pixels.  The function gets it's codes from
' get_next_code() which is responsible for reading blocks of data and
' separating them into the proper size codes.  Finally, GetByte() is
' the global routine to read the next byte from the GIF file.
'
' It is generally a good idea to have linewidth correspond to the actual
' width of a line (as specified in the Image header) to make your own
' code a bit simpler, but it isn't absolutely necessary.

Private Sub decoder(ByVal linewidth As Integer, buf() As Byte, stack() As Byte, suffix() As Byte, prefix() As Integer)
'assumed: all LBound's = 0
Dim sp As Integer, bufptr As Integer
Dim code As Integer, fc As Integer, oc As Integer, bufcnt As Integer
Dim c As Integer, size As Integer, ret As Integer

' Initialize for decoding a new image...
   size = GetByte
   If (size < 2) Or (9 < size) Then Error 321 'Invalid file format
   init_exp (size)

   ' Initialize in case they forgot to put in a clear code.
   ' (This shouldn't happen, but we'll try and decode it anyway...)
   oc = 0: fc = 0


   ' Set up the stack pointer and decode buffer pointer
   sp = 0
   bufptr = 0
   bufcnt = linewidth

   ' This is the main loop.  For each code we get we pass through the
   ' linked list of prefix codes, pushing the corresponding 'character' for
   ' each code onto the stack.  When the list reaches a single 'character'
   ' we push that on the stack too, and then start unstacking each
   ' character for output in the correct order.  Special handling is
   ' included for the clear code, and the whole thing ends when we get
   ' an ending code.
   
   Do
      c = get_next_code
      If c = ending Then Exit Do

      ' If the code is a clear code, reinitialize all necessary items.
      If c = clear Then
         curr_size = size + 1
         slot = newcodes
         top_slot = powers_of_two(curr_size)

         ' Continue reading codes until we get a non-clear code
         ' (Another unlikely, but possible case...)
         Do
            c = get_next_code
         Loop While c = clear

         ' If we get an ending code immediately after a clear code
         ' (Yet another unlikely case), then break out of the loop.
         If c = ending Then Exit Do

         ' Finally, if the code is beyond the range of already set codes,
         ' (This one had better NOT happen...  I have no idea what will
         ' result from this, but I doubt it will look good...) then set it
         ' to color zero.
         If c >= slot Then c = 0

         oc = c: fc = c

         ' And let us not forget to put the char into the buffer... And
         ' if, on the off chance, we were exactly one pixel from the end
         ' of the line, we have to send the buffer to the gif_out_line()
         ' routine...
         buf(bufptr) = c: bufptr = bufptr + 1
         bufcnt = bufcnt - 1
         If bufcnt = 0 Then
            OutLine buf, linewidth
            bufptr = 0
            bufcnt = linewidth
         End If
      Else

         ' In this case, it's not a clear code or an ending code, so
         ' it must be a code code...  So we can now decode the code into
         ' a stack of character codes. (Clear as mud, right?)
         code = c

         ' Here we go again with one of those off chances...  If, on the
         ' off chance, the code we got is beyond the range of those already
         ' set up (Another thing which had better NOT happen...) we trick
         ' the decoder into thinking it actually got the last code read.
         ' (Hmmn... I'm not sure why this works...  But it does...)
         If code >= slot Then
            If code > slot Then Error 321 'Invalid file format
            code = oc
            stack(sp) = fc: sp = sp + 1
         End If

         ' Here we scan back along the linked list of prefixes, pushing
         ' helpless characters (ie. suffixes) onto the stack as we do so.
         While code >= newcodes
            stack(sp) = suffix(code): sp = sp + 1
            code = prefix(code)
         Wend

         ' Push the last character on the stack, and set up the new
         ' prefix and suffix, and if the required slot number is greater
         ' than that allowed by the current bit size, increase the bit
         ' size.  (NOTE - If we are all full, we *don't* save the new
         ' suffix and prefix...  I'm not certain if this is correct...
         ' it might be more proper to overwrite the last code...
         
         stack(sp) = code: sp = sp + 1
         If slot < top_slot Then
            suffix(slot) = code: fc = code
            prefix(slot) = oc: slot = slot + 1
            oc = c
         End If
         If slot >= top_slot Then
            If curr_size < 12 Then
               top_slot = top_slot * 2
               curr_size = curr_size + 1
            End If
         End If
         ' Now that we've pushed the decoded string (in reverse order)
         ' onto the stack, lets pop it off and put it into our decode
         ' buffer...  And when the decode buffer is full, write another
         ' line...
         While sp > 0
            sp = sp - 1
            buf(bufptr) = stack(sp)
            bufptr = bufptr + 1
            bufcnt = bufcnt - 1
            If bufcnt = 0 Then
               If OutLine(buf, linewidth) Then Exit Sub
               bufptr = 0
               bufcnt = linewidth
            End If
         Wend
      End If
   Loop
   If bufcnt <> linewidth Then OutLine buf, linewidth - bufcnt
End Sub

Public Sub LoadFile(ByVal Filename As String)
Dim FileNumber As Integer
    FileNumber = FreeFile
    Open Filename For Binary Access Read As FileNumber
    ReDim data(LOF(1) - 1)
    Get FileNumber, , data()
    Close FileNumber
End Sub

Public Sub Reset()
    pdata = 0: Zombie = False
    
Dim gh As gif_header ', start As Long
    
'    Get FileNumber, , gh
    CopyMemory gh, data(pdata), Len(gh): pdata = pdata + Len(gh)
    Width = gh.Width
    Height = gh.Height
    
    If gh.giftype <> "GIF87a" And _
        gh.giftype <> "GIF89a" Then Error 321 'Invalid file format
    bih.biSize = Len(bih)
    bih.biWidth = gh.Width
    bih.biHeight = -gh.Height
    bih.biPlanes = 1
    bih.biBitCount = 32
    bih.biSizeImage = 4& * gh.Height * gh.Width
    
    ReDim BitmapData(0 To gh.Width - 1, 0 To gh.Height - 1)
    ReDim MaskData(0 To gh.Width - 1, 0 To gh.Height - 1)
    
    If (gh.flags And &H80) Then
        GlobalColors = 2 ^ ((gh.flags And &H7) + 1)
'        start = Loc(FileNumber)
'        Get FileNumber, , GlobalPalette
'        Seek FileNumber, 1 + start + 3 * GlobalColors
        CopyMemory GlobalPalette(0), data(pdata), 3 * GlobalColors
        pdata = pdata + 3 * GlobalColors

        GetMem4 ByVal VarPtrArray(GlobalPalette()), ByVal VarPtrArray(Palette())
        
''        Dim filler As RGBQUAD
''        With Palette(gh.bgcolor)
''            filler.rgbRed = .rgbRed
''            filler.rgbGreen = .rgbGreen
''            filler.rgbBlue = .rgbBlue
''        End With
''        GetMem4 filler, BackColor
''        FillMemory ByVal pvData, bih.biSizeImage, lFiller
        ZeroMemory MaskData(0, 0), bih.biSizeImage
    End If
End Sub

Public Function NextFrame() As StdPicture
Static gi As gif_image, dispos As Byte
Dim LocalColors As Integer, LocalPalette(0 To 255) As RGBTRIPLE
Dim Sig As Byte ', start As Long
    If Zombie Then Exit Function
    
    If dispos = 2 Then 'other disposal methods are ignored
        Dim i As Integer
        For i = gi.top To gi.top + gi.Height - 1
''            FillMemory BitmapData(gi.left, i), 4 * gi.width, lFiller
            ZeroMemory MaskData(gi.left, i), 4 * gi.Width
        Next
    End If
    
    GetMem4 ByVal VarPtrArray(GlobalPalette()), ByVal VarPtrArray(Palette())
    XprntIndex = -1: DelayMS = 0
    
    Sig = GetByte
    While Sig = &H21
        Dim gc As gif_control
        Sig = GetByte
        If Sig = &HF9 Then
'            Get FileNumber, , gc
            CopyMemory gc, data(pdata), Len(gc): pdata = pdata + Len(gc)
            If gc.size <> 4 Then Error 321 'Invalid file format
            If gc.terminator <> 0 Then Error 321  'Invalid file format
            If gc.flags And 1 Then
                XprntIndex = gc.xprntcolor
            End If
            DelayMS = 10 * gc.delay
            dispos = (gc.flags And &H1C) \ 4
            'user input reqmnt are ignored
        Else
            Do
                Sig = GetByte
'                Seek FileNumber, Loc(FileNumber) + Sig + 1
                pdata = pdata + Sig
            Loop Until Sig = 0
        End If
        Sig = GetByte
    Wend
    If Sig = &H3B Then
        Set NextMask = Nothing
        Zombie = True
        Exit Function
    End If
    If Sig <> &H2C Then Error 321 'Invalid file format
'    Get FileNumber, , gi
    CopyMemory gi, data(pdata), Len(gi): pdata = pdata + Len(gi)
    
    LineNo = gi.top: LastLine = LineNo + gi.Height - 1
    LineOff = gi.left
    If (gi.flags And &H80) Then
        LocalColors = 2 ^ ((gi.flags And &H7) + 1)
'        start = Loc(FileNumber)
'        Get FileNumber, , LocalPalette
'        Seek FileNumber, 1 + start + 3 * LocalColors
        CopyMemory LocalPalette(0), data(pdata), 3 * LocalColors
        pdata = pdata + 3 * LocalColors
        
        GetMem4 ByVal VarPtrArray(LocalPalette()), ByVal VarPtrArray(Palette())
    End If
    
    If ArrayExists(VarPtrArray(Palette())) = 0 Then Error 321 'Invalid file format
    
    Dim buf() As Byte, stack() As Byte, suffix() As Byte, prefix() As Integer
    ReDim buf(0 To gi.Width)
    ReDim stack(0 To MAX_CODES)
    ReDim suffix(0 To MAX_CODES)
    ReDim prefix(0 To MAX_CODES)
    decoder gi.Width, buf, stack, suffix, prefix
    If GetByte() <> 0 Then Error 321 'Invalid file format
    
Dim hBmp As Long, hMask As Long, hDC As Long: hDC = GetDC(0)
    hBmp = CreateDIBitmap(hDC, bih, CBM_INIT, BitmapData(0, 0), bih, 0)
'TODO: use monochrome bitmap here
    hMask = CreateDIBitmap(hDC, bih, CBM_INIT, MaskData(0, 0), bih, 0)
    Set NextFrame = CreatePictureFromBitmap(hBmp)
    Set NextMask = CreatePictureFromBitmap(hMask)
    ReleaseDC 0, hDC
    
    If ArrayExists(VarPtrArray(Palette())) Then PutMem4 ByVal VarPtrArray(Palette()), 0&
End Function

Private Function CreatePictureFromBitmap(ByVal hBmp As Long) As StdPicture
Dim Pic As PICTDESC, IID_IDispatch As Guid
    With IID_IDispatch
       .Data1 = &H20400
       .Data4(0) = &HC0
       .Data4(7) = &H46
    End With
    With Pic
       .size = Len(Pic)
       .Type = vbPicTypeBitmap
       .hBmp = hBmp
       .hPal = 0
    End With
    OleCreatePictureIndirect Pic, IID_IDispatch, 1, CreatePictureFromBitmap
End Function

Private Function ArrayExists(ByVal ppArr As Long) As Long ':-))
  GetMem4 ByVal ppArr, ArrayExists
End Function


Public Property Get GIFdata() As Byte()
    GIFdata = data
End Property

Public Property Let GIFdata(RHS() As Byte)
    data = RHS
End Property

